@@ -2,16 +2,17 @@
 
 test_that("mmrm_control works as expected", {
   result <- mmrm_control(
-    optimizer = stats::optim,
+    optimizer_fun = stats::optim,
     optimizer_args = list(method = "L-BFGS-B")
   )
   expected <- structure(
     list(
-      optimizer = stats::optim,
-      optimizer_args = list(method = "L-BFGS-B"),
-      optimizer_control = list(),
+      optimizers = result$optimizers,
       start = NULL,
-      accept_singular = TRUE
+      accept_singular = TRUE,
+      method = "Satterthwaite",
+      n_cores = 1L,
+      drop_visit_levels = TRUE
     ),
     class = "mmrm_control"
   )
@@ -214,8 +215,12 @@ test_that("h_mmrm_tmb_data works as expected", {
   formula <- FEV1 ~ RACE + us(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   result <- expect_silent(h_mmrm_tmb_data(
-    formula_parts, fev_data, fev_data$WEIGHT,
-    reml = FALSE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    fev_data$WEIGHT,
+    reml = FALSE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   ))
   expect_class(result, "mmrm_tmb_data")
   expect_named(
@@ -242,8 +247,12 @@ test_that("h_mmrm_tmb_data works as expected for grouped covariance", {
   formula <- FEV1 ~ RACE + us(AVISIT | ARMCD / USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   result <- expect_silent(h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    reml = FALSE, weights = rep(1, nrow(fev_data)), accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    reml = FALSE,
+    weights = rep(1, nrow(fev_data)),
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   ))
   expect_class(result, "mmrm_tmb_data")
   expect_named(
@@ -269,8 +278,12 @@ test_that("h_mmrm_tmb_data works as expected for mutli-dimensional spatial expon
   formula <- FEV1 ~ RACE + sp_exp(VISITN, VISITN2 | ARMCD / USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   result <- expect_silent(h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    reml = FALSE, weights = rep(1, nrow(fev_data)), accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    reml = FALSE,
+    weights = rep(1, nrow(fev_data)),
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   ))
   expect_class(result, "mmrm_tmb_data")
   expect_named(
@@ -298,12 +311,20 @@ test_that("h_mmrm_tmb_data works also for character ID variable", {
   dat <- fev_data
   dat$USUBJID <- as.character(dat$USUBJID) # nolint
   result <- expect_silent(h_mmrm_tmb_data(
-    formula_parts, dat,
-    weights = rep(1, nrow(dat)), reml = FALSE, accept_singular = FALSE
+    formula_parts,
+    dat,
+    weights = rep(1, nrow(dat)),
+    reml = FALSE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   ))
   expected <- expect_silent(h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = FALSE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = FALSE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   ))
   expect_identical(result, expected)
 })
@@ -313,8 +334,12 @@ test_that("h_mmrm_tmb_data correctly processes design matrix below full rank cor
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   dat <- fev_data[11:25, ]
   result <- expect_silent(h_mmrm_tmb_data(
-    formula_parts, dat,
-    weights = rep(1, nrow(dat)), reml = FALSE, accept_singular = TRUE
+    formula_parts,
+    dat,
+    weights = rep(1, nrow(dat)),
+    reml = FALSE,
+    accept_singular = TRUE,
+    drop_visit_levels = TRUE
   ))
   assert_true(qr(result$x_matrix)$rank == ncol(result$x_matrix))
   assert_true(sum(result$x_cols_aliased) == 2)
@@ -327,8 +352,12 @@ test_that("h_mmrm_tmb_data gives error for rank deficient design matrix when not
   dat <- fev_data[11:25, ]
   expect_error(
     h_mmrm_tmb_data(
-      formula_parts, dat,
-      weights = rep(1, nrow(dat)), reml = FALSE, accept_singular = FALSE
+      formula_parts,
+      dat,
+      weights = rep(1, nrow(dat)),
+      reml = FALSE,
+      accept_singular = FALSE,
+      drop_visit_levels = TRUE
     ),
     paste(
       "design matrix only has rank 8 and 2 columns (ARMCDTRT:AVISITVIS2, ARMCDTRT:AVISITVIS3)",
@@ -338,14 +367,90 @@ test_that("h_mmrm_tmb_data gives error for rank deficient design matrix when not
   )
 })
 
+test_that("h_mmrm_tmb_data catches case with multiple time points per subject early", {
+  formula_parts <- structure(
+    list(
+      formula = y ~ cd + ad(visit | id),
+      model_formula = y ~ cd,
+      full_formula = y ~ cd + id + visit,
+      cov_type = "ad",
+      is_spatial = FALSE,
+      visit_var = "visit",
+      subject_var = "id",
+      group_var = NULL
+    ),
+    class = "mmrm_tmb_formula_parts"
+  )
+
+  set.seed(123)
+  dat <- data.frame(
+    id = as.factor(c(1, 1, 1, 2, 2, 2)),
+    visit = as.factor(c(1, 2, 3, 2, 2, 3)),
+    cd = c(1, 2, 3, 4, 5, 6),
+    y = rnorm(6)
+  )
+
+  expect_error(
+    h_mmrm_tmb_data(
+      formula_parts,
+      data = dat,
+      weights = rep_len(1, nrow(dat)),
+      reml = TRUE,
+      accept_singular = TRUE,
+      drop_visit_levels = TRUE
+    ),
+    paste0(
+      "time points have to be unique for each subject, detected following duplicates in data:",
+      "\\s+id visit\\s+5\\s+2\\s+2" # Make sure we get the nice duplicates printed there.
+    )
+  )
+})
+
+
+test_that("h_mmrm_tmb_data has no side effect of overwrite the weights in global env", {
+  weights <- "this is global weights"
+  formula <- FEV1 ~ RACE + SEX + ARMCD * AVISIT + us(AVISIT | USUBJID)
+  formula_parts <- h_mmrm_tmb_formula_parts(formula)
+  res <- h_mmrm_tmb_data(
+    formula_parts,
+    data = fev_data,
+    weights = rep_len(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = TRUE,
+    drop_visit_levels = TRUE
+  )
+  expect_identical(weights, "this is global weights")
+})
+
+test_that("h_mmrm_tmb_data will not be affecte by `weights` in data", {
+  weights <- fev_data$WEIGHT
+  fev_data$weights <- weights
+  formula <- FEV1 ~ RACE + SEX + ARMCD * AVISIT + us(AVISIT | USUBJID)
+  formula_parts <- h_mmrm_tmb_formula_parts(formula)
+  res <- h_mmrm_tmb_data(
+    formula_parts,
+    data = fev_data,
+    weights = rep_len(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = TRUE,
+    drop_visit_levels = TRUE
+  )
+  expect_identical(res$weights_vector, rep_len(1, nrow(res$x_matrix)))
+  expect_identical(fev_data$weights, weights)
+})
+
 # h_mmrm_tmb_parameters ----
 
 test_that("h_mmrm_tmb_parameters works as expected without start values", {
   formula <- FEV1 ~ SEX + us(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 10))
@@ -356,8 +461,12 @@ test_that("h_mmrm_tmb_parameters works as expected with start values", {
   formula <- FEV1 ~ SEX + us(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   start <- 1:10
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = start))
@@ -369,8 +478,12 @@ test_that("h_mmrm_tmb_parameters works as expected with antedependence", {
   formula <- FEV1 ~ SEX + ad(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 4)) # 4 parameters.
@@ -381,8 +494,12 @@ test_that("h_mmrm_tmb_parameters works as expected with heterogeneous antedepend
   formula <- FEV1 ~ SEX + adh(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 7)) # 2 * 4 - 1 parameters.
@@ -393,8 +510,12 @@ test_that("h_mmrm_tmb_parameters works as expected with Toeplitz", {
   formula <- FEV1 ~ SEX + toep(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 4)) # 4 parameters.
@@ -405,8 +526,12 @@ test_that("h_mmrm_tmb_parameters works as expected with heterogeneous Toeplitz",
   formula <- FEV1 ~ SEX + toeph(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 7)) # 2 * 4 - 1 parameters.
@@ -417,8 +542,12 @@ test_that("h_mmrm_tmb_parameters works as expected with autoregressive", {
   formula <- FEV1 ~ SEX + ar1(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = c(0, 0.5))
@@ -429,8 +558,12 @@ test_that("h_mmrm_tmb_parameters works as expected with heterogeneous autoregres
   formula <- FEV1 ~ SEX + ar1h(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = c(rep(0, 4), 0.5)) # 4 + 1 parameters.
@@ -441,8 +574,12 @@ test_that("h_mmrm_tmb_parameters works as expected with compound symmetry", {
   formula <- FEV1 ~ SEX + cs(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 2))
@@ -453,8 +590,12 @@ test_that("h_mmrm_tmb_parameters works as expected with heterogeneous compound s
   formula <- FEV1 ~ SEX + csh(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    weights = rep(1, nrow(fev_data)), reml = TRUE, accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    weights = rep(1, nrow(fev_data)),
+    reml = TRUE,
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 5)) # 4 + 1 parameters.
@@ -469,7 +610,8 @@ test_that("h_mmrm_tmb_parameters works as expected with spatial exponential", {
     fev_data,
     reml = TRUE,
     accept_singular = FALSE,
-    weights = rep(1, nrow(fev_data))
+    weights = rep(1, nrow(fev_data)),
+    drop_visit_levels = TRUE
   )
   result <- expect_silent(h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL))
   expected <- list(theta = rep(0, 2)) # 1 + 1 parameters.
@@ -569,8 +711,12 @@ test_that("h_mmrm_tmb_extract_cov works as expected", {
   formula <- FEV1 ~ RACE + us(AVISIT | USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    reml = FALSE, weights = rep(1, nrow(fev_data)), accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    reml = FALSE,
+    weights = rep(1, nrow(fev_data)),
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   tmb_parameters <- h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL)
   tmb_object <- TMB::MakeADFun(
@@ -603,8 +749,12 @@ test_that("h_mmrm_tmb_extract_cov works as expected for group covariance", {
   formula <- FEV1 ~ RACE + ar1(AVISIT | ARMCD / USUBJID)
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    reml = FALSE, weights = rep(1, nrow(fev_data)), accept_singular = FALSE
+    formula_parts,
+    fev_data,
+    reml = FALSE,
+    weights = rep(1, nrow(fev_data)),
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   tmb_parameters <- h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL, n_groups = 2L)
   tmb_object <- TMB::MakeADFun(
@@ -648,7 +798,8 @@ test_that("h_mmrm_tmb_fit works as expected", {
     fev_data,
     weights = weights,
     reml = FALSE,
-    accept_singular = FALSE
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   tmb_parameters <- h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL)
   tmb_object <- TMB::MakeADFun(
@@ -699,8 +850,12 @@ test_that("h_mmrm_tmb_fit works as expected for grouped covariance", {
   formula_parts <- h_mmrm_tmb_formula_parts(formula)
   weights <- rep(1, nrow(fev_data))
   tmb_data <- h_mmrm_tmb_data(
-    formula_parts, fev_data,
-    reml = TRUE, weights = weights, accept_singular = TRUE
+    formula_parts,
+    fev_data,
+    reml = TRUE,
+    weights = weights,
+    accept_singular = TRUE,
+    drop_visit_levels = TRUE
   )
   tmb_parameters <- h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL, n_group = tmb_data$n_groups)
   tmb_object <- TMB::MakeADFun(
@@ -752,7 +907,8 @@ test_that("h_mmrm_tmb_fit errors when an invalid covariance type is used", {
     fev_data,
     weights = rep(1, nrow(fev_data)),
     reml = FALSE,
-    accept_singular = FALSE
+    accept_singular = FALSE,
+    drop_visit_levels = TRUE
   )
   tmb_parameters <- h_mmrm_tmb_parameters(formula_parts, tmb_data, start = NULL)
 
@@ -1035,7 +1191,7 @@ test_that("h_mmrm_tmb works with grouped toeph covariance structure and ML", {
   # PBO covariance matrix.
   expected_pbo_var <- c(104.38, 40.7267, 24.9011, 127.01)
   result_pbo_var <- as.numeric(diag(cor_mat$PBO))
-  expect_equal(result_pbo_var, expected_pbo_var, tolerance = 1e-4)
+  expect_equal(result_pbo_var, expected_pbo_var, tolerance = 1e-3)
   expected_pbo_rho <- c(0.3296, -0.1196, -0.3575)
   result_pbo_rho <- map_to_cor(result$theta_est[5:7])
   expect_equal(result_pbo_rho, expected_pbo_rho, tolerance = 1e-2)
@@ -1062,14 +1218,14 @@ test_that("h_mmrm_tmb works with grouped toeph covariance structure and REML", {
   # PBO covariance matrix.
   expected_pbo_var <- c(104.43, 40.8152, 25.0541, 127.27)
   result_pbo_var <- as.numeric(diag(cor_mat$PBO))
-  expect_equal(result_pbo_var, expected_pbo_var, tolerance = 1e-4)
+  expect_equal(result_pbo_var, expected_pbo_var, tolerance = 1e-3)
   expected_pbo_rho <- c(0.3315, -0.1172, -0.3567)
   result_pbo_rho <- map_to_cor(result$theta_est[5:7])
   expect_equal(result_pbo_rho, expected_pbo_rho, tolerance = 1e-2)
   # TRT covariance matrix.
   expected_trt_var <- c(74.6153, 34.2766, 49.8505, 220.20)
   result_trt_var <- as.numeric(diag(cor_mat$TRT))
-  expect_equal(result_trt_var, expected_trt_var, tolerance = 1e-4)
+  expect_equal(result_trt_var, expected_trt_var, tolerance = 1e-3)
   expected_trt_rho <- c(0.4677, 0.1540, -0.2741)
   result_trt_rho <- map_to_cor(result$theta_est[12:14])
   expect_equal(result_trt_rho, expected_trt_rho, tolerance = 1e-2)
@@ -1515,7 +1671,7 @@ test_that("h_mmrm_tmb works with sp_exp covariance structure and REML(2-dimensio
 
 ## misc ----
 
-test_that("h_mmrm_tmb also works with character ID variable", {
+test_that("fit_mmrm also works with character ID variable", {
   formula <- FEV1 ~ us(AVISIT | USUBJID)
   data <- fev_data
   data$USUBJID <- as.character(data$USUBJID) # nolint
@@ -1524,7 +1680,7 @@ test_that("h_mmrm_tmb also works with character ID variable", {
   expect_identical(result$beta_est, expected$beta_est)
 })
 
-test_that("h_mmrm_tmb saves data name in call element as expected", {
+test_that("fit_mmrm saves data name in call element as expected", {
   formula <- FEV1 ~ us(AVISIT | USUBJID)
   fit <- fit_mmrm(formula, fev_data, weights = rep(1, nrow(fev_data)))
   saved_call <- fit$call
@@ -1532,19 +1688,17 @@ test_that("h_mmrm_tmb saves data name in call element as expected", {
   expect_identical(saved_call$data, "fev_data")
 })
 
-test_that("h_mmrm_tmb works even when time point variable has unused factor levels", {
-  # Create a data set where one visit level only has NA in the data.
+test_that("fit_mmrm works even when time point variable has unused factor levels", {
   tmp_data <- fev_data
-  tmp_data$FEV1_BL[1] <- tmp_data$FEV1[1] <- NA # nolint
-  tmp_data$AVISIT <- as.character(tmp_data$AVISIT) # nolint
-  tmp_data$AVISIT[1] <- "SCREENING" # nolint
-  tmp_data$AVISIT <- as.factor(tmp_data$AVISIT) # nolint
-
-  result <- expect_silent(fit_mmrm(
-    FEV1 ~ FEV1_BL + RACE + us(AVISIT | USUBJID),
-    data = tmp_data,
-    weights = rep(1, nrow(tmp_data))
-  ))
+  tmp_data$AVISIT <- factor(tmp_data$AVISIT, levels = c("SCREENING", "VIS1", "VIS2", "VIS3", "VIS4"))
+  expect_message(
+    result <- fit_mmrm(
+      FEV1 ~ FEV1_BL + RACE + us(AVISIT | USUBJID),
+      data = tmp_data,
+      weights = rep(1, nrow(tmp_data))
+    ),
+    "In AVISIT there are dropped visits: SCREENING"
+  )
   expect_class(result, "mmrm_tmb")
   expect_identical(
     rownames(VarCorr(result)),
@@ -1552,7 +1706,44 @@ test_that("h_mmrm_tmb works even when time point variable has unused factor leve
   )
 })
 
-test_that("h_mmrm_tmb works with below full rank original design matrix by default", {
+test_that("fit_mmrm works if we keep the unused factor levels for specific covariance structure", {
+  tmp_data <- fev_data
+  tmp_data$AVISIT <- factor(tmp_data$AVISIT, levels = c("SCREENING", "VIS1", "VIS2", "VIS3", "VIS4"))
+
+  expect_silent(result <- fit_mmrm(
+    FEV1 ~ FEV1_BL + RACE + ar1(AVISIT | USUBJID),
+    data = tmp_data,
+    weights = rep(1, nrow(tmp_data)),
+    control = mmrm_control(drop_visit_levels = FALSE)
+  ))
+  expect_class(result, "mmrm_tmb")
+  expect_identical(
+    rownames(VarCorr(result)),
+    c("SCREENING", "VIS1", "VIS2", "VIS3", "VIS4")
+  )
+})
+
+test_that("fit_mmrm warns if we keep the unused factor levels for unstructured covariance", {
+  tmp_data <- fev_data
+  tmp_data$AVISIT <- factor(tmp_data$AVISIT, levels = c("SCREENING", "VIS1", "VIS2", "VIS3", "VIS4"))
+
+  expect_warning(
+    result <- fit_mmrm(
+      FEV1 ~ FEV1_BL + RACE + us(AVISIT | USUBJID),
+      data = tmp_data,
+      weights = rep(1, nrow(tmp_data)),
+      control = mmrm_control(drop_visit_levels = FALSE)
+    ),
+    "Model convergence problem: hessian is singular, theta_vcov not available"
+  )
+  expect_class(result, "mmrm_tmb")
+  expect_identical(
+    rownames(VarCorr(result)),
+    c("SCREENING", "VIS1", "VIS2", "VIS3", "VIS4")
+  )
+})
+
+test_that("fit_mmrm works with below full rank original design matrix by default", {
   formula <- FEV1 ~ RACE + SEX + SEX2 + ARMCD * AVISIT + us(AVISIT | USUBJID)
   dat <- fev_data
   dat$SEX2 <- dat$SEX # nolint
